from datetime import datetime
import os
import numpy as np
import pandas as pd
import json

from django.template.defaultfilters import slugify
from django.core import management
from django.core.management.base import BaseCommand
from django.conf import settings

class Command(BaseCommand):
    help = 'Generates Django models from a spreadsheet file.'

    def add_arguments(self, parser):
        parser.add_argument('filename', nargs='+', type=str)
        parser.add_argument('app', nargs='+', type=str)
        parser.add_argument('--overwrite', action='store_true', help='Overwrite existing models.py and admin.py files?')
        parser.add_argument('--migrate', action='store_true', help='Run makemigrations and migrate when done?')
        parser.add_argument('--loaddata', action='store_true', help='Load data after migrations are complete?')

    def handle(self, *args, **options):

        filename = options['filename'][0]
        app = options['app'][0]
        overwrite = options['overwrite']
        migrate = options['migrate']
        loaddata = options['loaddata']

        self.stdout.write(f'Converting {filename} to models in {app}')

        def field_name(column_name):
            return slugify(column_name).replace('-', '_')

        dtype_field_mapping = {
            'object': 'models.CharField({})',
            'bool': 'models.BooleanField({})',
            'int64': 'models.IntegerField({})',
            'float64': 'models.DecimalField({})',
        }

        def column_format_kwargs(series, dtype):

            kwargs = {}

            if str(dtype) == 'object':
                kwargs = {'max_length': max([len(str(cell_value)) for cell_value in series])}

            if str(dtype) == 'float64':
                def num_digits_and_precision(value:str) -> tuple:
                    total_digits = len(value.replace('.', ''))
                    dot = value.find('.')
                    decimals = value[dot+1:]
                    decimal_places = len(decimals)

                    return total_digits, decimal_places

                all_num_digits_and_precision = [num_digits_and_precision(str(cell_value)) for cell_value in series]
                max_digits = max([n for n, _ in all_num_digits_and_precision])
                decimal_places = max([n for _, n in all_num_digits_and_precision])

                kwargs = {'max_digits': max_digits, 'decimal_places': decimal_places}

            # series_has_null_values = any([cell_value == '' for cell_value in series])
            # series_has_null_values = random([True, False, False, False, False])
            series_has_null_values = False
            if series_has_null_values:
                kwargs['null'] = True
                kwargs['blank'] = True

            return kwargs

        def kwargs_string(kwargs):
            return ', '.join(f'{k}={v}' for k,v in kwargs.items())

        df = pd.read_excel(filename)

        # first rename the dataframe columns to their slugified equivalents
        for column in df.columns:
            df.rename(columns={column: field_name(column)}, inplace=True)

        df_dict = df.to_dict(orient='records')

        model_name = 'ConvertedModel'





        column_dtypes = [(column, df.dtypes[i]) for i, column in enumerate(df.columns)]

        fields = [
            (
                column,
                dtype_field_mapping[str(dtype)].format(kwargs_string(column_format_kwargs(df[column], dtype)))
            )
            for column, dtype
            in column_dtypes
        ]

        models_py = (f'''
# Created by django-from-excel at {datetime.now()}

from django.db import models

class {model_name}(models.Model):
    {"""
    """.join([" = ".join(field) for field in fields])}

    __str__ = __repr__ = lambda self: f\'{{self.id}}\'

    def __str__(self):
        return f\'{{self.id}}\'

        ''')

        admin_py = f'''
# Created by django-from-excel at {datetime.now()}

from django.contrib import admin
from .models import *

admin.site.register({model_name})

'''




        suffix ='' if overwrite else f'_{hex(abs(hash(datetime.now())))[2:10]}'
        models_py_filepath = os.path.join(app, f'models{suffix}.py')

        with open(models_py_filepath, 'w') as f:
            f.write(models_py)
        
        admin_py_filepath = os.path.join(app, f'admin{suffix}.py')

        with open(admin_py_filepath, 'w') as f:
            f.write(admin_py)

        self.stdout.write(self.style.SUCCESS(f'Generated {admin_py_filepath} and {models_py_filepath}'))

        if overwrite and migrate:
            management.call_command('makemigrations')
            management.call_command('migrate')

            if loaddata:
                fixture = [
                    {
                        'model': f'{app}.{model_name.lower()}',
                        'pk': i,
                        'fields': fields,
                    }
                    for i, fields
                    in enumerate(df_dict, start=1)
                ]

                fixtures_directory = os.path.join(settings.BASE_DIR, app, 'fixtures')
                fixtures_filename = os.path.join(fixtures_directory, f'{model_name.lower()}.json')
                if not os.path.exists(fixtures_directory):
                    os.makedirs(fixtures_directory)
                with open(fixtures_filename, 'w') as f:
                    json.dump(fixture, f)

                management.call_command('loaddata', f'{model_name.lower()}.json', app=app)

        self.stdout.write(self.style.SUCCESS(f'Complete. View the data at http://localhost:8000/admin/'))












# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_from_excel',
 'django_from_excel.management',
 'django_from_excel.management.commands']

package_data = \
{'': ['*']}

install_requires = \
['Django>=3.2.7,<4.0.0', 'openpyxl>=3.0.8,<4.0.0', 'pandas>=1.3.3,<2.0.0']

setup_kwargs = {
    'name': 'django-from-excel',
    'version': '0.1.1',
    'description': 'Automatically builds Django models from an Excel file',
    'long_description': "# django-from-excel\nAutomatically builds Django models from an Excel file\n\n## Getting Started\n\nThe following instructions assume you already have a Django project and at least one app set up.\n\n### Installation\n\n``` sh\npip install django-from-excel\n```\n\n### Setup\n\nCopy the Excel file you want to convert to a location in the project that is easy to reference. The simplest approach is to place it in the same directory as `manage.py`, as that is where you'll need to run the command from.\n\nBefore continuing, make sure you have created an app, and that it is registered in `settings.INSTALLED_APPS`.\n\nAdd `django_from_excel` to `settings.INSTALLED_APPS`.\n\nAt this point, also confirm your database settings.\n\n### Run\n\nExample:\n\n``` sh\nmanage.py buildfrom tracker.xlsx app --overwrite --migrate --loaddata\n  ```\n\nThe following options are available for the `buildfrom` command:\n\n**filename:**\n\nRequired. The filename or path/to/filename for the Excel file to generate models from.\n\n**app:**\n\nRequired. The app to generate `models.py` and `admin.py` files in. Must already exist, and must be listed in `settings.INSTALLED_APPS`.\n\n**--overwrite:**\n\nWill overwrite existing `admin.py` and `models.py` files in the target app; otherwise, a unique hex string will be appended to the filenames, such as `models_120f77f8.py`.\n\nWARNING: If you have existing models and specify `--overwrite`, they will be lost forever!\n\n**--migrate:**\n\nAfter generating the models, this will automatically run `manage.py makemigrations` and `manage.py migrate`.\n\n**--loaddata:**\n\nOnce migrations are complete, this will generate a `convertedmodel.json` file in `<app>/fixtures`, and will then call the `manage.py loaddata convertedmodel.json <app>` command.\nMust be used in conjunction with `--migrate`.\n\n### Inspect\n\nView the `models.py` and `admin.py` files that were generated.\n\nIf you chose the `--migrate` option:\n\nLog in to the [Django admin](http://localhost:8000/admin/). Your converted model(s) should be listed there.\n\nIf you specified `--loaddata`:\n\nOpen the model, and you should have a record for each row in the original Excel file.\n\n### Build Your Dream App\n\nThis is only intended to create a starting point. You'll want to inspect the models to make sure the field types are correct, you'll likely need to add some nullable fields or unique constraints, and you may need to spin some fields off into their own tables, with foreign keys linking back. But, hopefully `django-from-excel` saved you time by quickly creating some Django models for you.\n\n### Eject\n\nOnce you are satisfied with the results, you can remove `django_from_excel` from `settings.INSTALLED_APPS` and uninstall `django-from-excel` and any other dependencies you don't need for your project.\n\n---\n\n## TODO's\n\n1. Detect nullable data by columns with blank cells, while maintaining ability to detect column's data type\n2. Detect duplicate data that should be in a foreign table, and create corresponding ForeignKey field\n3. Create an eject command\n\n\n\n\n\n",
    'author': 'John Macy',
    'author_email': 'johncmacy@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/johncmacy/django-from-excel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
